/***************************************************************************
                netcdf.cpp  -  netCDF file data source reader
                             -------------------
    begin                : 17/06/2004
    copyright            : (C) 2004 Nicolas Brisset <nicodev@users.sourceforge.net>
    email                : kst@kde.org
    modified             : 03/14/05 by K. Scott

    modified/hacked      : 2013-04-09 by Patrick Rix
                            -->  support for PACKED / COMPRESSED DATA in short integer format
                            -->  plotting COMPONENTS of time dependent MATRICES  X_(time,j_comp)  as 1D-fields or vectors x_j(time)

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// ===============================================================================
//  P.Rix: support for plotting of  PACKED / COMPRESSED DATA  for variables having attributes:  'scale_factor' & 'add_offset'
// ===============================================================================
//  For variables of DataType  'ncSHORT'  automatic detection of scaling attributes
//  'scale_factor' and 'add_offset' used for data compression (float,double --> short int)
//  and for back-transformation (short int --> double) has been implemented.
//
//  If a variable owns the special attributes  ‘scale_factor’  and  ‘add_offset’  these
//  indicate that the variable has been packed/compressed from float/double(4/8 byte) to
//  (2 byte, signed) short integer data format. Each packed variable has a its own set of
//  scaling attributes  ‘varname:scale_factor’  and  ‘varname:add_offset’ saved in the
//  NetCDF file along with the variable's data. For the packing convention refer to:
//  <http://www.unidata.ucar.edu/software/netcdf/docs/netcdf/Attribute-Conventions.html>
//
//  The original floating point value ‘X(t_i)’ of a time dependent variable ‘X(t)’ at time index ‘t_i’
//  can be retrieved back by applying its scaling attributes ‘X:scale_factor’ and ‘X:add_offset’ to the
//  short integer values ‘I_short(t)’ as follows:
//
//         X(t_i) = X:scale_fator * I_short(t_i)  +  X:add_offset
// ===============================================================================
//
//
// ===============================================================================
//  P.Rix: HACK for plotting COMPONENTS(columns/rows) of time dependent MATRICES :  X_[time,j_comp]
// ===============================================================================
//  A netcdf file can contains 2D-matrix-data which could be plotted by KST (over INDEX)
//  along one dimension (row or column) while the index of the other dimension is kept fix.
//  Especially 2D-data of time-dependent matrices of the form  X_[time,j_comp]  could be
//  treated as a set of  j  time-dependent 1D-vectors or fields of the form x_1,2,3,..,j(time).
//  In order to plot the time series of certain components of the matrix in a XY-plot the
//  existing netcdf interface has to be enhanced to offer access to each single component
//  of a (time-dep.) matrix within the data wizard's dialog when the user has to choose a certain
//  data set for reading.
//  The  IDEA / STRATEGY  is to add the matrix' components as pseudo vector variables (fiels)
//  in the list of available data fields by taking the name of the matrix var as basename with
//  an extra ".[j]" index string (counting from 1) appended at the end and read in the data for
//  a chosen component from the corresponding matrix(!) vaiable using the appropriate array
//  index j-1 .
//  Offering a matrix-component in the list of available fields while reading the data from a
//  matrix var needs a little bit of tweaking the interface where the appended ".[j]" index string
//  serves as flag indicating that this vector var is a pseudo field not really contained in the
//  original netcdf data file.
// ===============================================================================




// TODO move
//#define KST_DEBUG_SHARED

#include "sharedptr.h"

#include "netcdfsource.h"
#include "kst_i18n.h"
#include "debug.h"

#include <QFile>
#include <QFileInfo>

#include <ctype.h>
#include <stdlib.h>


using namespace Kst;

static const QString netCdfTypeString = I18N_NOOP("netCDF Files");

int parse_netCDFvarName ( QString varName, bool& IS_MATRIX, QString& matrixVarName, int& j_matrixComp  )
{
    int success =  0;
    int failure = -1;

    int result = failure;

    // initialize assuming no matrix
    IS_MATRIX = false;
    matrixVarName = "";
    j_matrixComp = -1;

    // KST_DBG qDebug() << " function parse_netCDFvarName(): parsing var " << varName << " for '.[ ]' index string" << endl;

    int p1 = varName.lastIndexOf(".[");
    int p2 = varName.lastIndexOf("]");
    // KST_DBG qDebug() << " p1= " << p1 <<  " , p2= " << p2 << endl;

    if( (p1 >= 0)&&(p2 >= 0)&&( p2-(p1+2) > 0) ) {

        matrixVarName = varName;
        matrixVarName.truncate( p1 ); // cut off ".[j]" index str to get matrix var name

        int z;
        char j_str[15];    for(z=0; z<15; z++){ j_str[z] = '\0'; }
        z=0;
        for(int p=p1+2; p<p2; p++){
            if( (strlen(j_str) == 0)&&(varName[p] == QChar('0') )) {
                // skip leading zeros --> do nothing !
            }else{
                //j_str[z] = varName[p];
                j_str[z] = varName.toStdString().c_str()[p];
                z++;
            }
        }
        QString str = j_str;
        bool OK;
        z = str.toInt( &OK );
        if( OK ) {
            if( z > 0 ) {
                j_matrixComp = z-1; // index j needs count from  0  to access particular component of a matrix
                IS_MATRIX = true;

                //matrixVar = _ncfile->get_var( matrixVarName.toStdString().c_str() );
                //if (!matrixVar) {
                //  KST_DBG qDebug() << "Queried matrix " << matrixVarName << " cannot be read" << endl;
                //  return -1;
                //}
            }
        }

        if( IS_MATRIX ) {
            KST_DBG qDebug() << " ..var " << varName << " derived from matrix-var :  " << matrixVarName << " , j-comp.: " << j_matrixComp << endl;
        }

    }
  result = success;

  return result;

} // end parse_netCDFvarName()



//
// Scalar interface
//

class DataInterfaceNetCdfScalar : public DataSource::DataInterface<DataScalar>
{
public:
  DataInterfaceNetCdfScalar(NetcdfSource& s) : netcdf(s) {}

  // read one element
  int read(const QString&, DataScalar::ReadInfo&);

  // named elements
  QStringList list() const { return netcdf._scalarList; }
  bool isListComplete() const { return true; }
  bool isValid(const QString&) const;

  // T specific
  const DataScalar::DataInfo dataInfo(const QString&) const { return DataScalar::DataInfo(); }
  void setDataInfo(const QString&, const DataScalar::DataInfo&) {}

  // meta data
  QMap<QString, double> metaScalars(const QString&) { return QMap<QString, double>(); }
  QMap<QString, QString> metaStrings(const QString&) { return QMap<QString, QString>(); }


private:
  NetcdfSource& netcdf;
};


int DataInterfaceNetCdfScalar::read(const QString& scalar, DataScalar::ReadInfo& p)
{
  return netcdf.readScalar(p.value, scalar);
}


bool DataInterfaceNetCdfScalar::isValid(const QString& scalar) const
{
  return  netcdf._scalarList.contains( scalar );
}


//
// String interface
//

class DataInterfaceNetCdfString : public DataSource::DataInterface<DataString>
{
public:
  DataInterfaceNetCdfString(NetcdfSource& s) : netcdf(s) {}

  // read one element
  int read(const QString&, DataString::ReadInfo&);

  // named elements
  QStringList list() const { return netcdf._strings.keys(); }
  bool isListComplete() const { return true; }
  bool isValid(const QString&) const;

  // T specific
  const DataString::DataInfo dataInfo(const QString&) const { return DataString::DataInfo(); }
  void setDataInfo(const QString&, const DataString::DataInfo&) {}

  // meta data
  QMap<QString, double> metaScalars(const QString&) { return QMap<QString, double>(); }
  QMap<QString, QString> metaStrings(const QString&) { return QMap<QString, QString>(); }


private:
  NetcdfSource& netcdf;
};


//-------------------------------------------------------------------------------------------
int DataInterfaceNetCdfString::read(const QString& string, DataString::ReadInfo& p)
{
  //return netcdf.readString(p.value, string);
  if (isValid(string) && p.value) {
    *p.value = netcdf._strings[string];
    return 1;
  }
  return 0;
}


bool DataInterfaceNetCdfString::isValid(const QString& string) const
{
  return netcdf._strings.contains( string );
}





//
// Vector interface
//

class DataInterfaceNetCdfVector : public DataSource::DataInterface<DataVector>
{
public:
  DataInterfaceNetCdfVector(NetcdfSource& s) : netcdf(s) {}

  // read one element
  int read(const QString&, DataVector::ReadInfo&);

  // named elements
  QStringList list() const { return netcdf._fieldList; }
  bool isListComplete() const { return true; }
  bool isValid(const QString&) const;

  // T specific
  const DataVector::DataInfo dataInfo(const QString&) const;
  void setDataInfo(const QString&, const DataVector::DataInfo&) {}

  // meta data
  QMap<QString, double> metaScalars(const QString&);
  QMap<QString, QString> metaStrings(const QString&);


private:
  NetcdfSource& netcdf;
};


const DataVector::DataInfo DataInterfaceNetCdfVector::dataInfo(const QString &field) const
{
  if (!netcdf._fieldList.contains(field))
    return DataVector::DataInfo();

          /*
              bool IS_MATRIX = false;
              QString matrixVarName = "";
              int j_matrixComp = -1;

            parse_netCDFvarName ( field, IS_MATRIX, matrixVarName, j_matrixComp  );

            if( IS_MATRIX ) {
                NcVar *var = netcdf._ncfile->get_var( matrixVarName.toStdString().c_str() );
                int NT = var->get_dim(0)->size();

                return DataVector::DataInfo( NT, netcdf.samplesPerFrame(field));
            } else {

                return DataVector::DataInfo(netcdf.frameCount(field), netcdf.samplesPerFrame(field));
            }
          */

  return DataVector::DataInfo(netcdf.frameCount(field), netcdf.samplesPerFrame(field));
}



int DataInterfaceNetCdfVector::read(const QString& field, DataVector::ReadInfo& p)
{
  return netcdf.readField(p.data, field, p.startingFrame, p.numberOfFrames);
}


bool DataInterfaceNetCdfVector::isValid(const QString& field) const
{
  return  netcdf._fieldList.contains( field );
}

QMap<QString, double> DataInterfaceNetCdfVector::metaScalars(const QString& field)
{    
    bool IS_MATRIX = false;
    QString matrixVarName = "";
    int j_matrixComp = -1;

  parse_netCDFvarName ( field, IS_MATRIX, matrixVarName, j_matrixComp  );

  QString varName;

  if( IS_MATRIX ) {
      varName = matrixVarName;
  } else {
      varName =  field;
  }

  NcVar *var = netcdf._ncfile->get_var( varName.toStdString().c_str() );

  if (!var) {
      if( IS_MATRIX ) {
            KST_DBG qDebug() << "Queried field (matrix)" << field << " which can't be read <-- NetcdfSource::samplesPerFrame()" << endl;
      } else {
            KST_DBG qDebug() << "Queried field (vector)" << field << " which can't be read <-- NetcdfSource::samplesPerFrame()" << endl;
      }
      return QMap<QString, double>();
  }

  /*
      NcVar *var = netcdf._ncfile->get_var(field.toLatin1().constData());
      if (!var) {
        KST_DBG qDebug() << "Queried field " << field << " which can't be read <-- DataInterfaceNetCdfVector::metaScalars()" << endl;
        return QMap<QString, double>();
      }
  */

      bool HAS_ATT__SCALE_FACTOR = false;
      bool HAS_ATT__ADD_OFFSET   = false;
      double scale_factor = 0.0;
      double add_offset  = 0.0;

  QMap<QString, double> fieldScalars;

  fieldScalars["NbAttributes"] = var->num_atts();
  for (int i=0; i<var->num_atts(); ++i) {
    NcAtt *att = var->get_att(i);
    QString AttName = att->name() ;

    if( AttName.toLower() == "scale_factor" ) {
        HAS_ATT__SCALE_FACTOR = true;
        scale_factor = att->values()->as_double(0);
    }

    if( AttName.toLower() == "add_offset" ) {
        HAS_ATT__ADD_OFFSET = true;
        add_offset = att->values()->as_double(0);
    }

    // Only handle char attributes as fieldStrings, the others as fieldScalars
    if (att->type() == ncByte || att->type() == ncShort || att->type() == ncInt
        || att->type() == ncLong || att->type() == ncFloat || att->type() == ncDouble) {
      // Some attributes may have multiple values => load the first as is, and for the others
      // add a -2, -3, etc... suffix as obviously we can have only one value per scalar.
      // Do it in two steps to avoid a test in the loop while keeping a "clean" name for the first one
      fieldScalars[QString(att->name())] = att->values()->as_double(0);
      for (int j=1; j<att->values()->num(); ++j) {
        fieldScalars[QString(att->name()) + QString("-") + QString::number(j+1)] = att->values()->as_double(j);
      }
    }
  }
  return fieldScalars;
}

QMap<QString, QString> DataInterfaceNetCdfVector::metaStrings(const QString& field)
{
    bool IS_MATRIX = false;
    QString matrixVarName = "";
    int j_matrixComp = -1;

  parse_netCDFvarName ( field, IS_MATRIX, matrixVarName, j_matrixComp  );

  QString varName;

  if( IS_MATRIX ) {
      varName = matrixVarName;
  } else {
      varName =  field;
  }

  NcVar *var = netcdf._ncfile->get_var( varName.toStdString().c_str() );

  if (!var) {
      if( IS_MATRIX ) {
            KST_DBG qDebug() << "Queried field (matrix)" << field << " which can't be read <-- NetcdfSource::samplesPerFrame()" << endl;
      } else {
            KST_DBG qDebug() << "Queried field (vector)" << field << " which can't be read <-- NetcdfSource::samplesPerFrame()" << endl;
      }
      return QMap<QString, QString>();
  }

/*
      NcVar *var = netcdf._ncfile->get_var(field.toLatin1().constData());
      if (!var) {
        KST_DBG qDebug() << "Queried field " << field << " which can't be read <-- DataInterfaceNetCdfVector::metaStrings()" << endl;
        return QMap<QString, QString>();
      }
*/

  QMap<QString, QString> fieldStrings;

  QString tmpString;
  for (int i=0; i<var->num_atts(); ++i) {
    NcAtt *att = var->get_att(i);
    // Only handle char/unspecified attributes as fieldStrings, the others as fieldScalars
    if (att->type() == ncChar || att->type() == ncNoType) {
      fieldStrings[att->name()] = QString(att->values()->as_string(0));
    }
    // qDebug() << att->name() << ": " << att->values()->num() << endl;
  }
  return fieldStrings;
}


//
// Matrix interface
//

class DataInterfaceNetCdfMatrix : public DataSource::DataInterface<DataMatrix>
{
public:

  DataInterfaceNetCdfMatrix(NetcdfSource& s) : netcdf(s) {}

  // read one element
  int read(const QString&, DataMatrix::ReadInfo&);

  // named elements
  QStringList list() const { return netcdf._matrixList; }
  bool isListComplete() const { return true; }
  bool isValid(const QString&) const;

  // T specific
  const DataMatrix::DataInfo dataInfo	(const QString&) const;
  void setDataInfo(const QString&, const DataMatrix::DataInfo&) {}

  // meta data
  QMap<QString, double> metaScalars(const QString&) { return QMap<QString, double>(); }
  QMap<QString, QString> metaStrings(const QString&) { return QMap<QString, QString>(); }


private:
  NetcdfSource& netcdf;
};


const DataMatrix::DataInfo DataInterfaceNetCdfMatrix::dataInfo(const QString& matrix) const
{
  if (!netcdf._matrixList.contains( matrix ) ) {
    return DataMatrix::DataInfo();
  }

  QByteArray bytes = matrix.toLatin1();
  NcVar *var = netcdf._ncfile->get_var(bytes.constData());  // var is owned by _ncfile
  if (!var) {
    return DataMatrix::DataInfo();
  }

  if (var->num_dims() != 2) {
    return DataMatrix::DataInfo();
  }

  DataMatrix::DataInfo info;
  info.samplesPerFrame = 1;
  // TODO is this right?
  info.xSize = var->get_dim(0)->size();
  info.ySize = var->get_dim(1)->size();

  return info;
}


int DataInterfaceNetCdfMatrix::read(const QString& field, DataMatrix::ReadInfo& p)
{
  int count = netcdf.readMatrix(p.data->z, field);

  p.data->xMin = 0;
  p.data->yMin = 0;
  p.data->xStepSize = 1;
  p.data->yStepSize = 1;

  return count;
}


bool DataInterfaceNetCdfMatrix::isValid(const QString& field) const {
  return  netcdf._matrixList.contains( field );
}


//
// NetcdfSource
//

NetcdfSource::NetcdfSource(Kst::ObjectStore *store, QSettings *cfg, const QString& filename, const QString& type, const QDomElement &element) :
  Kst::DataSource(store, cfg, filename, type),
  _ncfile(0L),
  _ncErr(NcError::silent_nonfatal),
  is(new DataInterfaceNetCdfScalar(*this)),
  it(new DataInterfaceNetCdfString(*this)),
  iv(new DataInterfaceNetCdfVector(*this)),
  im(new DataInterfaceNetCdfMatrix(*this))
  {
  setInterface(is);
  setInterface(it);
  setInterface(iv);
  setInterface(im);

  setUpdateType(None);

  if (!type.isEmpty() && type != "netCDF") {
    return;
  }

  _valid = false;
  _maxFrameCount = 0;

  _filename = filename;
  _strings = fileMetas();
  _valid = initFile();
}


NetcdfSource::~NetcdfSource() {
  delete _ncfile;
  _ncfile = 0L;
}


void NetcdfSource::reset() {
  delete _ncfile;
  _ncfile = 0L;
  _maxFrameCount = 0;
  _valid = initFile();
}


bool NetcdfSource::initFile() {
  _ncfile = new NcFile(_filename.toUtf8().data(), NcFile::ReadOnly);
  if (!_ncfile->is_valid()) {
      qDebug() << _filename << ": failed to open in initFile()" << endl;
      return false;
    }

  KST_DBG qDebug() << _filename << ": building field list" << endl;
  _fieldList.clear();
  _fieldList += "INDEX";

  int nb_vars = _ncfile->num_vars();
  KST_DBG qDebug() << nb_vars << " vars found in total" << endl;

  _maxFrameCount = 0;

  for (int i = 0; i < nb_vars; i++) {
    NcVar *var = _ncfile->get_var(i);
    if (!var) {
      continue;
    }
    if (var->num_dims() == 0) {
      _scalarList += var->name();
    } else if (var->num_dims() == 1) {
      _fieldList += var->name();
      int fc = var->num_vals() / var->rec_size();
      _maxFrameCount = qMax(_maxFrameCount, fc);
      _frameCounts[var->name()] = fc;
    } else if (var->num_dims() == 2) {
      _matrixList += var->name();
    } // end if (var->num_dims() == 2) -->  reading _matrixList
  } // end for (int i = 0; i < nb_vars; i++)

/**/
  int nb_vectorVars_added = 0;

  for (int i = 0; i < nb_vars; i++) {
    NcVar *var = _ncfile->get_var(i);
    if (!var) {
      continue;
    }
    if (var->num_dims() == 0) {
      //_scalarList += var->name();
    } else if (var->num_dims() == 1) {
      //_fieldList += var->name();
      //int fc = var->num_vals() / var->rec_size();
      //_maxFrameCount = qMax(_maxFrameCount, fc);
      //_frameCounts[var->name()] = fc;
    } else if (var->num_dims() == 2) {
      _matrixList += var->name();
      
   // ===============================================================================
   //  P.Rix: HACK for plotting COMPONENTS(columns/rows) of time dependent MATRICES X[time,j_comp]
   // ===============================================================================
      // IDEA / STRATEGY:
      // check: is matrix time dependent ? 
      //        i.e. --> does matrix have a dimension named 'time' as 1st dimension ?
      //
        // if YES, then step along the 2nd dimension of the matrix and read in each component
        // (i.e. column or row) of the matrix as a separate vector of length time.
        // ..to do this, for each component of the matrix we have to:
        //
        //    + define & create (i.e. add) a new (pseudo) vector variable of length 'time'
        //      after the original vars, i.e.:
        //       * derive vector-var-name from matrix-var-name by appending the current component index ".[j]" along 2nd dimension
        //       * use identical data type for vector var as was used for matrix var
        //    + add the new (pseudo) vector variable to '_fieldList'
        //      and set '_maxFrameCount' & '_frameCounts' appropriately        
   // ===============================================================================             

      bool IS_TIME_DEPENDENT = false;
                  
      long    dim0Size = var->get_dim(0)->size();
      NcToken dim0Name = var->get_dim(0)->name();
      //QString Dim0Name = "dimension_0_name" ;
      QString Dim0Name = (const char*) dim0Name ;
      
      long NT = 0;
      NcDim*  dimTime = NULL;
      if( Dim0Name.toLower() == "time" ){
      	 IS_TIME_DEPENDENT = true; 
      	 dimTime = var->get_dim(0);
      	 NT = var->get_dim(0)->size();
      }
      
      if( IS_TIME_DEPENDENT ) {

      	long    dim1Size = var->get_dim(1)->size();
      	NcToken dim1Name = var->get_dim(1)->name();
        //QString Dim1Name = "dimension_1_name" ;
        QString Dim1Name = (const char*) dim1Name ;

      	NcType dataType  = var->type(); 
      	
        QString matrixVarName = (const char*) var->name() ;
        QString vectorVarName = "" ;
        QString vectorVarBaseName = matrixVarName.append( '.' ); // take matrix-var-name as vector base name and append '.' as delimiter


        KST_DBG qDebug() << " ..adding " << dim1Size << " new 1D-vector-vars of length " << NT << " for matrix-var :  " << matrixVarName << endl;

        for(int j = 0; j < dim1Size; j++) {

            char j_str[15];    j_str[0] = '\0';

            sprintf(j_str, "[%02d]", j+1); // add +1 to index j to count from 1; index string for vals < 10: pad a leading zero on the left.

            vectorVarName = vectorVarBaseName;
            vectorVarName.append( j_str ); // append index-string "[j]" to vector base name
      		
            NcVar* vectorVar = _ncfile->add_var( vectorVarName.toStdString().c_str(), dataType, dimTime ); // add new vector var

            nb_vectorVars_added++;

            //_fieldList += vectorVar->name();
            _fieldList += vectorVarName; // add vector-var-name to list of plottable 1D-variables

            //int fc = vectorVar->num_vals() / var->rec_size();
            int fc = NT;
            _maxFrameCount = qMax(_maxFrameCount, fc);
            _frameCounts[ vectorVarName ] = fc;
            //_frameCounts[ vectorVar->name( )] = fc;

            KST_DBG qDebug() << " vector-var added: " << vectorVarName << endl;
      		
      	} // for j
      } // end if( IS_TIME_DEPENDENT )		
   // ===============================================================================

    } // end if (var->num_dims() == 2) -->  reading _matrixList
  } // end for (int i = 0; i < nb_vars; i++)

  KST_DBG qDebug() << nb_vectorVars_added << " vector-vars added in total (from 2nd matrix dims)" << endl;
/**/


  // Get strings
  int globalAttributesNb = _ncfile->num_atts();
  for (int i = 0; i < globalAttributesNb; ++i) {
    // Get only first value, should be enough for a start especially as strings are complete
    NcAtt *att = _ncfile->get_att(i);
    if (att) {
      QString attrName = QString(att->name());
      char *attString = att->as_string(0);
      QString attrValue = QString(att->as_string(0));
      delete[] attString;
      //TODO port
      //KstString *ms = new KstString(KstObjectTag(attrName, tag()), this, attrValue);
      _strings[attrName] = attrValue;
    }
    delete att;
  }

  // TODO update(); // necessary?  slows down initial loading
  return true;
} // end NetcdfSource::initFile()



Kst::Object::UpdateType NetcdfSource::internalDataSourceUpdate() {
  //TODO port
  /*
  if (KstObject::checkUpdateCounter(u)) {
    return lastUpdateResult();
  }
  */

  _ncfile->sync();

  bool updated = false;
  /* Update member variables _ncfile, _maxFrameCount, and _frameCounts
     and indicate that an update is needed */
  int nb_vars = _ncfile->num_vars();
  for (int j = 0; j < nb_vars; j++) {
    NcVar *var = _ncfile->get_var(j);
    if (!var) {
      continue;
    }
    int fc = var->num_vals() / var->rec_size();
    _maxFrameCount = qMax(_maxFrameCount, fc);
    updated = updated || (_frameCounts[var->name()] != fc);
    _frameCounts[var->name()] = fc;
  }
  return updated ? Object::Updated : Object::NoChange;
}


int NetcdfSource::readScalar(double *v, const QString& field)
{
  // TODO error handling
  QByteArray bytes = field.toLatin1();
  NcVar *var = _ncfile->get_var(bytes.constData());  // var is owned by _ncfile
  if (var) {
    var->get(v);
    return 1;
  }
  return 0;
}

int NetcdfSource::readString(QString *stringValue, const QString& stringName)
{
  // TODO more error handling?
  NcAtt *att = _ncfile->get_att((NcToken) stringName.toLatin1().data());
  if (att) {
    *stringValue = QString(att->as_string(0));
    delete att;
    return 1;
  }
  return 0;
}

int NetcdfSource::readField(double *v, const QString& field, int s, int n) {
  NcType dataType = ncNoType; /* netCDF data type */
  /* Values for one record */
  NcValues *record = 0;// = new NcValues(dataType,numFrameVals);


  KST_DBG qDebug() << "Entering NetcdfSource::readField with params: " << field << ", from " << s << " for " << n << " frames" << endl;

  /* For INDEX field */
  if (field.toLower() == "index") {
    if (n < 0) {
      v[0] = double(s);
      return 1;
    }
    for (int i = 0; i < n; ++i) {
      v[i] = double(s + i);
    }
    return n;
  }


  KST_DBG qDebug() << "Query field " << field << endl;

  /* For a variable from the netCDF file */
/*
  QByteArray bytes = field.toLatin1();
  NcVar *var = _ncfile->get_var(bytes.constData());  // var is owned by _ncfile

  if (!var) {
    KST_DBG qDebug() << "Queried field " << field << " which can't be read" << endl;
    return -1;
  }

 // dataType = var->type();
*/

  // ===============================================================================
  //  P.Rix: HACK for plotting COMPONENTS(columns/rows) of time dependent MATRICES
  // ===============================================================================
     // IDEA / STRATEGY:
     // during initialisation (i.e. opening) of the netcdf file it was already checked if the file contains
     // time dependent matrix variables (with 'time' as 1st dimension) and new (pseudo) vector variables should
     // have been added for each matrix component with an index string ".[j]" appended to the (matrix var) base name
     // indicating that this vector var was artifically added.
     //
     //   --> check the variable's name for the presence of a ".[j]" index string at the end:
     //         if an index str was found, read the data of component j-1 of the corresponding MATRIX var,
     //         otherwise read data from a normal vector or field var
     //
  // ===============================================================================

  //QString varName = (const char*) var->name() ; // == field name
  QString varName = field ;


  bool READ_MATRIX_COMPONENT = false;

    bool IS_MATRIX = false;
    QString matrixVarName = "";
    int j_matrixComp = -1;

  parse_netCDFvarName ( field, IS_MATRIX, matrixVarName, j_matrixComp  );


  if( IS_MATRIX ) {

      READ_MATRIX_COMPONENT = true;

      NcVar  *matrixVar = _ncfile->get_var( matrixVarName.toStdString().c_str() );
      if (!matrixVar) {
        KST_DBG qDebug() << "Queried matrix " << matrixVarName << " cannot be read" << endl;
        return -1;
      }

      dataType = matrixVar->type();

      int NT = matrixVar->get_dim(0)->size(); // length of vector/field == number of time steps

      KST_DBG qDebug() << " ..query 1D-vector-data of length " << NT << " from matrix-var :  " << matrixVarName << " , comp.: " << j_matrixComp << endl;

      bool OK_type   = true;
      bool OK_length = true;
      bool OK_index  = true;

      //OK_type   = (var->type() == matrixVar->type()) ;
      //OK_length = (var->get_dim(0)->size() == matrixVar->get_dim(0)->size()) ;
      //OK_index  = ( (0 <= j_matrixComp)&&(j_matrixComp < matrixVar->get_dim(1)->size())) ;

      int N_read = 0;

      if(OK_type && OK_length && OK_index ) {

          if (s >= NT) {
            return 0;
          }
          N_read = n;
          if( s+N_read > NT ) { N_read = NT-s;  };

          switch (dataType) {
          case ncShort:
          {
              // ..do we have a COMPRESSED VARIABLE ?
              //    --> check the presence of attributes  scale_factor  &  add_offset
                          bool HAS_ATT__SCALE_FACTOR = false;
                          bool HAS_ATT__ADD_OFFSET   = false;
                          double scale_factor = 0.0;
                          double add_offset  = 0.0;

                        for (int i=0; i<matrixVar->num_atts(); ++i) {
                            NcAtt *att = matrixVar->get_att(i);
                            QString AttName = att->name() ;

                            if( AttName.toLower() == "scale_factor" ) {
                                HAS_ATT__SCALE_FACTOR = true;
                                scale_factor = att->values()->as_double(0);
                            }
                            if( AttName.toLower() == "add_offset" ) {
                                HAS_ATT__ADD_OFFSET = true;
                                add_offset = att->values()->as_double(0);
                            }
                        }


                      short* vals = (short*) malloc ( sizeof( short )*NT );
                      matrixVar->set_cur(0, j_matrixComp);
                      matrixVar->get( &vals[0], NT, 1);
                      for (int i = 0; i < N_read; i++) {
                          if( HAS_ATT__SCALE_FACTOR && HAS_ATT__ADD_OFFSET ) {
                              v[i] = scale_factor * vals[i]  +  add_offset;
                          } else {
                            v[i] = vals[i];
                          }
                      }
                      //vectorVar->put( vals );
                      free(vals); vals=NULL;

          }
          break;

          case ncInt:
          {
                      int* vals = (int*) malloc ( sizeof( int )*NT );
                      matrixVar->set_cur(0, j_matrixComp);
                      matrixVar->get( &vals[0], NT, 1);
                      for (int i = 0; i < N_read; i++) {
                          v[i] = vals[i];
                      }
                      //vectorVar->put( vals );
                      free(vals); vals=NULL;
          }
          break;

          case ncFloat:
          {
                      float* vals = (float*) malloc ( sizeof( float )*NT );
                      matrixVar->set_cur(0, j_matrixComp);
                      matrixVar->get( &vals[0], NT, 1);
                      for (int i = 0; i < N_read; i++) {
                          v[i] = vals[i];
                      }
                      //vectorVar->put( vals );
                      free(vals); vals=NULL;
          }
          break;

          case ncDouble:
          {
                      double* vals = (double*) malloc ( sizeof( double )*NT );
                      matrixVar->set_cur(0, j_matrixComp);
                      matrixVar->get( &vals[0], NT, 1);
                      for (int i = 0; i < N_read; i++) {
                          v[i] = vals[i];
                      }
                      //vectorVar->put( vals );
                      free(vals); vals=NULL;
          }
          break;

          default:
            KST_DBG qDebug() << field << ": wrong datatype for kst, no values read !" << endl;
            return -1;
            break;
          } // end switch (dataType)
      } // end if(OK_type && OK_length && OK_index ) {

      KST_DBG qDebug() << "Finished reading field " << j_matrixComp << " of matrix " << matrixVarName << endl;

    return N_read;

  } else{
      // =======================================================
      //   STANDARD CASE: read data from a 1D vector or field
      // =======================================================

      /* For a variable from the netCDF file */
      QByteArray bytes = field.toLatin1();
      NcVar *var = _ncfile->get_var(bytes.constData());  // var is owned by _ncfile

      if (!var) {
        KST_DBG qDebug() << "Queried field (vector)" << field << " which can't be read <-- NetcdfSource::readField()" << endl;
        return -1;
      }

      dataType = var->type();

      if (s >= var->num_vals() / var->rec_size()) {
        return 0;
      }

      bool oneSample = n < 0;
      int recSize = var->rec_size();

      switch (dataType) {
        case ncShort:
          {
          // ..do we have a COMPRESSED VARIABLE ?
          //    --> check the presence of attributes  scale_factor  &  add_offset
                      bool HAS_ATT__SCALE_FACTOR = false;
                      bool HAS_ATT__ADD_OFFSET   = false;
                      double scale_factor = 0.0;
                      double add_offset  = 0.0;

                    for (int i=0; i<var->num_atts(); ++i) {
                        NcAtt *att = var->get_att(i);
                        QString AttName = att->name() ;

                        if( AttName.toLower() == "scale_factor" ) {
                            HAS_ATT__SCALE_FACTOR = true;
                            scale_factor = att->values()->as_double(0);
                        }
                        if( AttName.toLower() == "add_offset" ) {
                            HAS_ATT__ADD_OFFSET = true;
                            add_offset = att->values()->as_double(0);
                        }
                    }


            if (oneSample) {
              record = var->get_rec(s);
              if( HAS_ATT__SCALE_FACTOR && HAS_ATT__ADD_OFFSET ) {
                  v[0] = scale_factor * record->as_short(0)  +  add_offset;
              } else {
                  v[0] = record->as_short(0);
              }
              delete record;
            } else {
              for (int i = 0; i < n; i++) {
                record = var->get_rec(i+s);
                for (int j = 0; j < recSize; j++) {
                    if ( HAS_ATT__SCALE_FACTOR && HAS_ATT__ADD_OFFSET ) {
                      v[i*recSize + j] = scale_factor * record->as_short(j)  +  add_offset;
                    } else {
                      v[i*recSize + j] = record->as_short(j);
                    }
                }
                delete record;
              }
            }
          }
          break;

        case ncInt:
          {
            if (oneSample) {
              record = var->get_rec(s);
              v[0] = record->as_int(0);
              delete record;
            } else {
              for (int i = 0; i < n; i++) {
                record = var->get_rec(i+s);
                KST_DBG qDebug() << "Read record " << i+s << endl;
                for (int j = 0; j < recSize; j++) {
                  v[i*recSize + j] = record->as_int(j);
                }
                delete record;
              }
            }
          }
          break;

        case ncFloat:
          {
            if (oneSample) {
              record = var->get_rec(s);
              v[0] = record->as_float(0);
              delete record;
            } else {
              for (int i = 0; i < n; i++) {
                record = var->get_rec(i+s);
                for (int j = 0; j < recSize; j++) {
                  v[i*recSize + j] = record->as_float(j);
                }
                delete record;
              }
            }
          }
          break;

        case ncDouble:
          {
            if (oneSample) {
              record = var->get_rec(s);
              v[0] = record->as_double(0);
              delete record;
            } else {
              for (int i = 0; i < n; i++) {
                record = var->get_rec(i+s);
                for (int j = 0; j < recSize; j++) {
                  v[i*recSize + j] = record->as_double(j);
                }
                delete record;
              }
            }
          }
          break;

        default:
          KST_DBG qDebug() << field << ": wrong datatype for kst, no values read" << endl;
          return -1;
          break;

      } // end switch dataType

      KST_DBG qDebug() << "Finished reading field " << field << endl;

      return oneSample ? 1 : n * recSize;

  } // end if( READ_MATRIX_COMPONENT )..else..

} // end NetcdfSource::readField()





int NetcdfSource::readMatrix(double *v, const QString& field) 
{
  /* For a variable from the netCDF file */
  QByteArray bytes = field.toLatin1();
  NcVar *var = _ncfile->get_var(bytes.constData());  // var is owned by _ncfile
  if (!var) {
    KST_DBG qDebug() << "Queried field (matrix) " << field << " which can't be read <-- NetcdfSource::readMatrix()" << endl;
    return -1;
  }

  int xSize = var->get_dim(0)->size();
  int ySize = var->get_dim(1)->size();

  var->get(v, xSize, ySize);

 
  return  xSize * ySize;
}





int NetcdfSource::samplesPerFrame(const QString& field) {

  if (field.toLower() == "index") {
    return 1;
  }

  QByteArray bytes = field.toLatin1();

  //NcVar *var = _ncfile->get_var(bytes.constData());

    bool IS_MATRIX = false;
    QString matrixVarName = "";
    int j_matrixComp = -1;

  parse_netCDFvarName ( field, IS_MATRIX, matrixVarName, j_matrixComp  );

  if( IS_MATRIX ) {
      NcVar *matrixVar = _ncfile->get_var( matrixVarName.toStdString().c_str() );
      if (!matrixVar) {
        KST_DBG qDebug() << "Queried field (matrix)" << field << " which can't be read <-- NetcdfSource::samplesPerFrame()" << endl;
        return 0;
      }
      return 1; // return  rec_size = 1  as the components of a time dependent matrix consist of time dep. 1D-vectors vars

  } else {
      NcVar *var = _ncfile->get_var(bytes.constData());
      if (!var) {
        KST_DBG qDebug() << "Queried field (vector)" << field << " which can't be read <-- NetcdfSource::samplesPerFrame()" << endl;
        return 0;
      }

      return var->rec_size(); // should be rec_size=1 for a time dependent 1D vector variable
  }
  //return var->rec_size();
}



int NetcdfSource::frameCount(const QString& field) const {
  if (field.isEmpty() || field.toLower() == "index") {
    return _maxFrameCount;
  } else {

        /* */

              bool IS_MATRIX = false;
              QString matrixVarName = "";
              int j_matrixComp = -1;

            parse_netCDFvarName ( field, IS_MATRIX, matrixVarName, j_matrixComp  );

            QString varName;

            if( IS_MATRIX ) {
                NcVar *var = _ncfile->get_var( matrixVarName.toStdString().c_str() );
                int NT = var->get_dim(0)->size();

                KST_DBG qDebug() << "Queried field (matrix) " << field << " has " << NT << " frames <-- NetcdfSource::frameCount()" << endl;

                return NT;

            } else {

                KST_DBG qDebug() << "Queried field (vector) " << field << " has " << _frameCounts[field] << " frames <-- NetcdfSource::frameCount()" << endl;

                return _frameCounts[field];
            }


       KST_DBG qDebug() << "Queried field " << field << " has " << _frameCounts[field] << " frames <-- NetcdfSource::frameCount()" << endl;

       return _frameCounts[field];
  }
}



QString NetcdfSource::fileType() const {
  return "netCDF";
}



bool NetcdfSource::isEmpty() const {
  return frameCount() < 1;
}



const QString& NetcdfSource::typeString() const
{
  return netCdfTypeString;
}


const QString NetcdfSource::netcdfTypeKey()
{
  return ::netCdfTypeString;
}
